@@ -3,7 +3,7 @@ use strict;
 
 use lib 'lib';
 use Gearman::Worker;
-use Storable qw( thaw );
+use Storable qw(thaw nfreeze);
 use Getopt::Long qw( GetOptions );
 
 GetOptions(
@@ -25,6 +25,7 @@ $worker->register_function(sum => sub {
 });
 
 $worker->register_function(fail => sub { undef });
+$worker->register_function(fail_die => sub { die 'test reason' });
 $worker->register_function(fail_exit => sub { exit 255 });
 
 $worker->register_function(sleep => sub { sleep $_[0]->arg });
@@ -43,6 +44,13 @@ $worker->register_function(echo_prefix => sub {
     join " from ", $_[0]->arg, $prefix;
 });
 
+$worker->register_function(echo_sleep => sub {
+    my($job) = @_;
+    $job->set_status(1, 1);
+    sleep 2; ## allow some time to read the status
+    join " from ", $_[0]->arg, $prefix;
+});
+
 
 $worker->register_function(long => sub {
     my($job) = @_;
@@ -50,9 +58,35 @@ $worker->register_function(long => sub {
     sleep 2;
     $job->set_status(100, 100);
     sleep 2;
+    return $job->arg;
 });
 
 my $nsig;
 $nsig = kill 'USR1', $notifypid if $notifypid;
 
-$worker->work while 1;
+my $work_exit = 0;
+
+$worker->register_function(work_exit => sub {
+    $work_exit = 1;
+});
+
+my ($is_idle, $last_job_time);
+
+$worker->register_function(check_stop_if => sub {
+    return nfreeze([$is_idle, $last_job_time]);
+});
+
+
+
+my $stop_if = sub {
+    ($is_idle, $last_job_time) = @_;
+
+    if ($work_exit) {
+        $work_exit = 0;
+        return 1;
+    }
+
+    return 0;
+};
+
+$worker->work(stop_if => $stop_if) while (1);
